/***************************************************
** Microsoft SQL Server 2000 Programming by Example
** ISBN 0-7897-2449-9
** Carlos Eduardo Rojas and Fernando G. Guerrero
** Copyright 2001 by QUE Corporation
** Source code. Listing 10.8, Version 1.0, 2001-04-25
****************************************************/

USE Northwind
GO

-----------------------------------------------------------
-- Generic function to convert an angle
-- into HMS format
-- @d_g indicates if the angle is measured in
-- sexagesimal degrees (CHAR(176) as the degrees symbol) or
-- centesimal degrees ('g', for grades)
-----------------------------------------------------------

CREATE FUNCTION dbo.ANG_HMS
(@angle float, @precision int = 0, @d_g CHAR(1))
RETURNS varchar(21)
AS
BEGIN

	-- Declare variables
	
	DECLARE @sign float
	DECLARE @ncircles int
	
	DECLARE @degrees int
	DECLARE @minutes int
	DECLARE @seconds float
	DECLARE @secInt int
	DECLARE @secDec int
	
	DECLARE @frac float
	
	DECLARE @hms varchar(20)
	
	-- Save the sign of the angle
	
	SET @sign = SIGN(@angle)
	
	-- Take out the sign of the angle
	-- to avoid calculation problems
	
	SET @angle = ABS(@angle)
	
	-- Extract the integer part as degrees
	
	SET @degrees = CAST(FLOOR(@angle) AS int)
	
	-- Count how many complete circles the angle has
	
	SET @ncircles = @degrees /
		CASE @d_g 
			WHEN 'g' THEN 400
			ELSE 360 
		END
	
	-- Convert the angle into an angle from the first circle
	
	SET @degrees = @degrees %
		CASE @d_g 
			WHEN 'g' THEN 400
			ELSE 360 
		END
	
	-- Extract the decimal part from the angle
	
	SET @frac = @angle - @degrees - (@ncircles *
		CASE @d_g 
			WHEN 'g' THEN 400
			ELSE 360 
		END
	)
	
	-- Extract minutes from the decimal part
	
	SET @minutes = FLOOR(@frac *
		CASE @d_g 
			WHEN 'g' THEN 100
			ELSE 60 
		END
	)
	
	-- Extract the number of seconds
	
	SET @seconds = (@frac *
		CASE @d_g 
			WHEN 'g' THEN 100
			ELSE 60 
		END
		- @minutes) *
			CASE @d_g 
				WHEN 'g' THEN 100
				ELSE 60 
			END
	
	-- Calculate the number of complete seconds
	
	SET @secInt = FLOOR(@seconds)
	
	-- Set a limit for the fraction of a second to 9 digits
	
	IF @precision > 9
		SET @precision = 9
	
	-- Extract the fraction of a second in the given precision
	
	SET @secDec = (@seconds - @secInt) * POWER(10, @precision)
	
	-- Start creating the resulting string with the sign
	-- only for negative numbers
	
	SET @hms = 
		CASE @sign
			WHEN -1 THEN '- '
			ELSE '' 
		END
	
	-- Create the HMS format
	
	SET @hms = @hms
		+ CAST(@degrees AS varchar(3))
			+ CHAR(176) + ' '
		+ CAST(@minutes AS varchar(2))
			+ 'm '
		+ CAST(@secInt AS varchar(2))
			+ 's'
		+ CAST(@secDec AS varchar(10))
	
	RETURN @hms
	
END
GO

----------------------------------
-- Function to convert angles
-- measured in centesimal degrees
-- into HMS format
----------------------------------

CREATE FUNCTION dbo.GRAD_HMS
(@angle float, @precision int = 0)
RETURNS varchar(21)
AS
BEGIN

	-- Call the ANG_HMS function with the 'g' format
	
	RETURN dbo.ANG_HMS(@angle, @precision, 'g')

END
GO

-----------------------------------
-- Function to convert angles
-- measured in sexagesimal degrees
-- into HMS format
-----------------------------------

CREATE FUNCTION dbo.DEG_HMS
(@angle float, @precision int = 0)
RETURNS varchar(21)
AS
BEGIN

	-- Call the ANG_HMS function with the CHAR(176) format
	
	RETURN dbo.ANG_HMS(@angle, @precision, CHAR(176))

END
GO

-----------------------------------------------------------
-- Generic function to convert an angle
-- from HMS format
-- @d_g indicates if the angle is measured in
-- sexagesimal degrees (CHAR(176) as the degrees symbol) or
-- centesimal degrees ('g', for grades)
-----------------------------------------------------------

CREATE FUNCTION dbo.HMS_ANG
(@hms varchar(22), @d_g char(1))
RETURNS float
AS
BEGIN

	-- Declare variables
	
	DECLARE @sign float
	
	DECLARE @pos0 int
	DECLARE @posg int
	DECLARE @posm int
	DECLARE @poss int
	
	DECLARE @degrees float
	DECLARE @angle float
	DECLARE @minutes float
	DECLARE @seconds float
	DECLARE @secInt float
	DECLARE @secDec float
	
	-- Extract the sign
	
	IF RIGHT(@hms, 1) = '-'
		SELECT @sign = -1, @pos0 = 2
	ELSE
		SELECT @sign = 1, @pos0 = 1
	
	-- Search for the position in the string
	-- of the character dividing degrees, minutes, and seconds
	
	SET @posg = CHARINDEX(@d_g, @hms)
	SET @posm = CHARINDEX('m', @hms)
	SET @poss = CHARINDEX('s', @hms)
	
	-- Extract the value of the degrees
	
	SET @degrees = SUBSTRING(@hms, @pos0, @posg - @pos0)
	
	-- Extract the value of the minutes
	
	SET @minutes = SUBSTRING(@hms, @posg + 1, @posm - @posg - 1)
	
	-- Extract the value of the seconds
	-- as integer and decimal part
	
	SET @secInt = SUBSTRING(@hms, @posm + 1, @poss - @posm - 1)
	SET @secDec = SUBSTRING(@hms, @poss + 1, len(@hms) - @poss)
		/ POWER(10.0, len(@hms) - @poss)
	
	-- Calculate the angle
	
	SET @angle = @sign * (
		@degrees +
		@minutes /
			CASE @d_g 
				WHEN 'g' THEN 100.0
				ELSE 60.0 
			END +
		(@secInt + @secdec) /
			CASE @d_g 
				WHEN 'g' THEN 10000.0
				ELSE 3600.0 
			END)
	
	-- Return the value
	
	RETURN @angle
	
END
GO

----------------------------------
-- Function to convert angles
-- measured in sexagesimal degrees
-- from HMS format
----------------------------------

CREATE FUNCTION dbo.HMS_DEG
(@hms varchar(22))
RETURNS float
AS
BEGIN

	-- Call the HMS_ANG function with the CHAR(176) format
	
	RETURN dbo.HMS_ANG(@hms, CHAR(176))

END
GO

----------------------------------
-- Function to convert angles
-- measured in centesimal degrees
-- from HMS format
----------------------------------

CREATE FUNCTION dbo.HMS_GRAD
(@hms varchar(22))
RETURNS float
AS
BEGIN

	-- Call the ANG_HMS function with the 'g' format
	
	RETURN (dbo.HMS_ANG(@hms, 'g'))

END

GO 
